
from tokenize import group
from turtle import update
from microkit_python_server_sdk.publisher import Publisher
from microkit_python_server_sdk.config_item import ConfigItem
from microkit_python_server_sdk.features_item import FeaturesItem
class Group:
    types = ['config', 'features']
    def __init__(self, type, values, ):
        if type not in Group.types:
            raise Exception('Group type is not valid')
        self.group = {}
        self.type = type
        self.change = Publisher()
        self.updated = False
        self.initialized = False
        self.update(values)
   
    def update (self, values):
        prev_value = self.value
        for key, item in values.items():
            if key in self.group:
                self.group[key].update(item)
            else:
                if self.initialized:
                    self.updated = True
                if "value" in item and not type(item['value']) is dict:
                    self.group[key] =  ConfigItem(item, key) if self.type is 'config' else FeaturesItem(item, key)
                    def published(current_value, prev_value):
                        self.updated = True
                    self.group[key].change.subscribe(published)
                else:
                    self.group[key] =  Group(self.type, item)
            setattr(self, key, self.group[key])

        if not self.initialized:
            self.initialized = True 
        if self.updated:
            self.change.publish(self.value ,prev_value)
            self.updated = False 

    @property
    def value (self):
        val = {}
        for key, item in self.group.items():
            val[key] = item.value
        return val

    


from email.policy import HTTP
import unittest
from conf import Conf
from unittest.mock import patch, MagicMock


class ConfTestCase(unittest.TestCase):
    def test_init_options(self):
        options = {
            'base_url': 'test url',
            'api_key': 'key',
            'port': '443',
            'http': True,
            'polling_interval': 1000,
        }
        client = Conf(options)

        self.assertEqual(client.get('base_url'), options['base_url'])
        self.assertEqual(client.get('api_key'), options['api_key'])
        self.assertEqual(client.get('port'), options['port'])
        self.assertEqual(client.get('http'), options['http'])
        self.assertEqual(client.get('polling_interval'), options['polling_interval'])
   
    def test_init_options_default(self):
        options = {
            'base_url': 'test url',
            'api_key': 'key',
            'port': '443',
            'http': False,
            'polling_interval': 1000,
        }
        client = Conf( {"api_key": options['api_key']})

        self.assertEqual(client.get('base_url'), 'sdk.microkit.app')
        self.assertEqual(client.get('api_key'), options['api_key'])
        self.assertEqual(client.get('port'), 443)
        self.assertEqual(client.get('http'), False)
        self.assertEqual(client.get('polling_interval'), 30000)
    
    def test_raise_if_api_does_not_exist(self):
        self.assertRaises(Exception, Conf)
    
    

    

if __name__ == '__main__':
    unittest.main()


import json
import urllib.error
import urllib.parse
import urllib.request
import ssl




class HttpClient:

    instance = None

    def __init__(self, config):
        # To do raise error if url is not http
        if HttpClient.instance:
            return HttpClient.instance
        self.config = config
        HttpClient.instance = self

    def get_instance():
        return HttpClient.instance

    def post(self, action, uri = None, post_data = None):

        data = {"options": {"service": self.config.get('service'), "action": action, "lang": "python"}}
        if post_data:
            data = merge_two_dicts(data, post_data)
        
        if self.config.get("key_vars_values"):
            data = merge_two_dicts(data, {"key_vars_values": self.config.get("key_vars_values")})

        # Dict to Json
        data = json.dumps(data)

        # Convert to String
        data = str(data)

        # Convert string to byte
        data = data.encode('utf-8')

        headers = {
             'Content-Type': 'application/json',
             'Authorization': "Bearer {}".format(self.config.get('api_key'))
        }
        http = 'http' if self.config.get('http') else 'https'
        base_url = self.config.get('base_url')
        port = self.config.get('port')
        version = self.config.get('version')
        url = '%s://%s:%s/%s/%s'%(http, base_url, port, version, uri if uri else '')
        
        ssl_context = ssl.SSLContext() if not self.config.get('http') else None
        req = urllib.request.Request(url, headers=headers, data=data)
        with urllib.request.urlopen(req, context=ssl_context) as response:
            return response.read()

def merge_two_dicts(x, y):
    z = x.copy()   
    z.update(y)    
    return z


import unittest
from microkit import Microkit
from network_controller import NetworkController
from conf import Conf
from unittest.mock import patch, MagicMock

data = {"config": "dc5+idVVGi00GpKlmkp5PyMHZ0AzpZuyl+0sfTVzYcFHCc\/8bMGvamSA3jfYEfJZJ6i6eMu4vHIvpqBF+\/LH4m9tO6Armcm4vSk9cM1rlrtypLA\/q+Ms1EUgXBWdoeJbZOz\/k8QvRu7VGz\/s9RtFPL3ouzMdKaFAj78hIMscQG5uv1ClZoLNXbLrAzO9sJPmyFS4GfUulgZ2\/sih3G7qNVaMclHcIVX6vtvz09pEF7CRH1SMZCXd\/ceh\/e1ifpeZk1g762TNF37NBIDFMf0f50dvpM0OAo+U8+WZtJASPriD4zPVvp5mGR+LdOrPYFFHpby+cyAqutWMo16a1qTIkPcQkt0XRcIGlANrCXjiCwZ8c+Vu2z3feLPurq1E1HTiTwPVB5zRuaVYstLqDmEkV3AZZI0IMAL\/nHKJFLGkFeWfVQNBlCwgeqrbSqLDh5Pn3m6\/N3MEiypSivKGjRsUSMsyIV2nrmPVUMvE3BMkXsTA0u0RYSMGCIQhqf1RbtypCfrnPIsJVA7+JSHfkMx3ZIA+1BPL9n+Hbvb4RQxPwXp+3XXkhCzKfdc2loWxMFCV\/NSi9XcGrFATLAYb9tz4\/aNhzXU+WWOLa3+RflLLSDHFKvz18pTOCNdmK4fujLv5mgH0KmNpWnndFRWY\/mJjuetBxov8NolSFrIXUcE\/MzqEc2IeB68Fv9AXMw1Uz62+Cq3rQVeTnTlattRNrtamWBluUMqA\/bZKIUP1LV\/gF8kSVSVzLWZ\/iHg1mr1YxYswkoVoKIpp3qBmkN6CyKtyHNmyRSxL\/d5fUfi5tUHmbZ\/1lMMtA0v3KC2up+uC245uup1V9H++6YHe392FaYyZe41C5A3V3vFMJiTC2VnSd\/ZbmjNLfwhPvUqcfwLApYwyXWjwPd2gNNM2lL+VVhsjTuCFNGZ+YccSKQXW4EJzUjg2YB2R\/MqFEYePL\/OyAv4K=="}
key = '3c20444b-6865-4dfb-bfee-0a0b7279572b-qPYBtP5lgv4f/TDOV17BZA=='
base_url = 'test'

class MicrokitTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        pass
    
    
    
    
    @patch('urllib.request.urlopen')
    def test_initializeKit_return_Microkit_instanc (self,mock_urlopen):
        cm = MagicMock()
        cm.getcode.return_value = 200
        cm.read.return_value = data
        cm.__enter__.return_value = cm
        mock_urlopen.return_value = cm

        microkit = Microkit.initialize_kit(key)
        self.assertIsInstance(microkit, Microkit)

    @patch('urllib.request.urlopen')
    def test_network_controller_instanc (self,mock_urlopen):
        cm = MagicMock()
        cm.getcode.return_value = 200
        cm.read.return_value = data
        cm.__enter__.return_value = cm
        mock_urlopen.return_value = cm

        microkit = Microkit.initialize_kit(key)
        self.assertIsInstance(microkit.network_controller, NetworkController)
    
    @patch('urllib.request.urlopen')
    def test_config_base_url_return_Microkit_instanc (self,mock_urlopen):
        cm = MagicMock()
        cm.getcode.return_value = 200
        cm.read.return_value = data
        cm.__enter__.return_value = cm
        mock_urlopen.return_value = cm
        
        microkit = Microkit.initialize_kit(key, {}, {"base_url": "test"})
        self.assertEqual(microkit.config.get('base_url'), 'test')
    
    @patch('urllib.request.urlopen')
    def test_config_kit_item_value (self,mock_urlopen):
        cm = MagicMock()
        cm.getcode.return_value = 200
        cm.read.return_value = data
        cm.__enter__.return_value = cm
        mock_urlopen.return_value = cm
        
        microkit = Microkit.initialize_kit(key, {}, {"base_url": "test"})
        self.assertEqual(microkit.config_kit.ports.configurations.value, '8020')
    
    # @patch('urllib.request.urlopen')
    # def test_initializeKit_call_post (self,mock_urlopen):
    #     cm = MagicMock()
    #     cm.getcode.return_value = 200
    #     cm.read.return_value = data
    #     cm.__enter__.return_value = cm
    #     mock_urlopen.return_value = cm
        
    #     microkit = Microkit.initialize_kit(key, {}, {"base_url": "test"})
    #     self.assertTrue(mock_urlopen.called)
    
    # @patch('urllib.request.urlopen')
    # def test_initializeKit_call_post (self,mock_urlopen):
    #     cm = MagicMock()
    #     cm.getcode.return_value = 200
    #     cm.read.return_value = data
    #     cm.__enter__.return_value = cm
    #     mock_urlopen.return_value = cm
        
       
    #     res = self.client.post()
    #     self.assertEqual(res, data)

    

if __name__ == '__main__':
    unittest.main()


from multiprocessing.sharedctypes import Value
import unittest
from group import Group
from publisher import Publisher
from unittest.mock import Mock
from config_item import ConfigItem

class GroupTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.value = {
            'item_1': {
                'value': 'value_1',
                'type': 'type_1'
                },
            'item_2' : {
                'value': 'value_2',
                'type': 'type_2'
                }
        }
        

    def test_group_prop(self):
        group = Group('config', self.value)
        self.assertTrue(isinstance(group.group, type({})))
    
    def test_type_prop(self):
        group = Group('config', self.value)
        self.assertEqual(group.type, 'config')
    
    def test_exception_if_type_not_valid(self):
        self.assertRaises(Exception, Group ,'not_valid', self.value)
            
    def test_update_values(self):
        group = Group('config', self.value)
        self.assertTrue(isinstance(group.item_1, ConfigItem))

    def test_change_prop(self):
        group = Group('config', self.value)
        self.assertTrue(isinstance(group.change, Publisher)) 
    
    # def test_change_called_if_children_is_changed(self):
    #     group = Group('config', self.value)
    #     f = Mock()
    #     group.change.subscribe(f)
    #     new_value = self.value.copy()
    #     new_value['item_1']['value'] = 'other_value'
    #     group.update(new_value)
        
    #     f.assert_called_once()

    def test_change_called_with_all_children_values(self):
        group = Group('config', self.value)
        f = Mock()
        group.change.subscribe(f)
        new_value = self.value.copy()
        new_value['item_1']['value'] = 'other_value'
        group.update(new_value)
        f.assert_called_once_with({
                'item_1': 'other_value',
                'item_2' : 'value_2'
            },
            {
                'item_1': 'value_1',
                'item_2' : 'value_2'
            })
    
    # def test_value(self):
    #     value = {'value': 'value', 'type': 'type'}
    #     item = Group(value)
    #     self.assertEqual(item.value, 'value')
    
    # def test_change_prop(self):
    #     value = {'value': 'value', 'type': 'type'}
    #     item = Group(value)
    #     self.assertTrue(isinstance(item.change, Publisher))
    
    # def test_change_is_called_when_value_is_change(self):
    #     value = {'value': 'value', 'type': 'type'}
    #     item = Group(value)
    #     test_dict = {}
        
    #     def func (val):
    #         test_dict['value'] = val

    #     item.change.subscribe(func)
    #     item.value = 'change'
    #     self.assertEqual(test_dict['value'], item.value)

    # def test_change_is_not_called_when_value_is_not_change(self):
    #     value = {'value': 'value', 'type': 'type'}
    #     item = Group(value)
    #     f = Mock()

    #     item.change.subscribe(f)
    #     item.value = 'value'
    #     f.assert_not_called()
    
    

    

if __name__ == '__main__':
    unittest.main()


from microkit_python_server_sdk.conf import Conf
from microkit_python_server_sdk.network_controller import NetworkController
from microkit_python_server_sdk.group import Group
from microkit_python_server_sdk.aes import AES
import base64
import json

class Microkit:
    
    instance = None

    def __init__(self, key,  user , options):
        if Microkit.instance:
            return Microkit.instance
        options['api_key'] = key
        self.config = Conf(options=options, user=user)
        self.network_controller = NetworkController(self.config)
        Microkit.instance = self
   

    @staticmethod
    def initialize_kit (key, user = {}, options = {}):
        if not Microkit.instance:

            key_size = len(key)
            secret_size = 24
            sdk_key = key[:key_size-secret_size-1]
            secret = key[-secret_size:]
            Microkit(sdk_key, user, options)
            res = Microkit.instance.network_controller.get_latest_data('init')
            res = json.loads(res.decode('utf-8'))
            
            Microkit.instance.config.set("key_vars_values", res["key_vars_values"])
            Microkit.instance.config_kit = Group('config', Microkit.decrypt(res['config'], secret) if 'config' in res and res['config'] != '' else {})
            Microkit.instance.features_kit = Group('features', res["features"] if 'features' in res and type(res["features"]) == dict  else {})
            if Microkit.instance.config.get('polling_on'):
                Microkit.instance.network_controller.start_update_interval()
                def callback (current_val, prev_val):
                    current_val = json.loads(current_val.decode('utf-8'))
                    Microkit.instance.config_kit.update(Microkit.decrypt(current_val['config'], secret) if 'config' in current_val and current_val['config'] != '' else {})
                    Microkit.instance.features_kit.update(current_val["features"] if 'features' in current_val and type(current_val["features"]) == dict  else {})
                    Microkit.instance.config.set("key_vars_values", res["key_vars_values"])
                Microkit.instance.network_controller.change.subscribe(callback)

        return Microkit.instance

    def decrypt(raw_text, secret):
        key = base64.b64decode(secret)
        iv = base64.b64decode(raw_text)[:16]
        text = base64.b64decode(raw_text)[16:]
    
        decrypt_value = AES(key).decrypt_cbc(text, iv)
        return json.loads(decrypt_value.decode('utf-8'))

    def close (self):
        self.network_controller.stop_update_interval()
        Microkit.instance = None

    def kit_ready():
        if Microkit.instance.features_kit and Microkit.instance.config_kit:
            return Microkit.instance
        else:
            return False


import unittest
from publisher import Publisher
from unittest.mock import Mock

class PublisherTestCase(unittest.TestCase):
    def test_subscribers_array(self):
        pub = Publisher()
        self.assertEqual(pub.subscribers, {})
    
    def test_insert_func_into_subscribers_array(self):
        pub = Publisher()
        def func ():
            pass
        subscription = pub.subscribe(func)
        self.assertEqual(pub.subscribers[subscription.key], func)
    
    def test_unsubscribe(self):
        pub = Publisher()
        def func ():
            pass
        subscription = pub.subscribe(func)
        subscription.unsubscribe()
        self.assertEqual(pub.subscribers, {})
    
    
    def test_call_func_on_publish(self):
        pub = Publisher()
        f = Mock()
        subscription = pub.subscribe(f)
        pub.publish('current', 'prev')
        f.assert_called_with('current', 'prev')

    def test_call_func_on_first_subscription(self):
        pub = Publisher()
        f = Mock()
        subscription = pub.subscribe(f)
        subscription.call()
        f.assert_called_once()
    
    

    

if __name__ == '__main__':
    unittest.main()


from pyexpat import features
from re import S
from microkit_python_server_sdk.item import Item
from microkit_python_server_sdk.http_client import HttpClient
import json
class FeaturesItem(Item):
   def __init__(self, value, name):
        self.targeting_groups = value["targeting_groups"]
        self.client = HttpClient.get_instance()
        super(FeaturesItem, self).__init__(value, name)
        
   def getValue (self, user):
      if not user or len(self.targeting_groups) == 0:
         return self.value
      else:
         features = {}
         features[self.name] = {"value": self.value, "type": self.type, "targeting_groups": self.targeting_groups}
         res = self.client.post("get_user_data", "get_feature_by_user",{"user": user,  "features": features})
         res = json.loads(res.decode('utf-8'))
         return res["features"][self.name]["value"] if self.name in res["features"] else self.value
    
    

    


import unittest
from microkit_python_server_sdk.microkit import Microkit




class MicrokitTestCase(unittest.TestCase):
   
    def test_initializeKit (self):
        microkit = Microkit.initialize_kit("83578f1b-e460-43ce-a481-a1efd5382496-B+XtrA4P/2nDYHmrh+Ts/g==", {}, {"base_url": "localhost", "port": "8030", "http": True, 'polling_on': False})
        print(microkit.features_kit.value)
        def callback(current, prev):
            print(current)
            print(prev)
            
        microkit.config_kit.ports.change.subscribe(callback)

    

if __name__ == '__main__':
    unittest.main()


import unittest
from config_item import ConfigItem
from publisher import Publisher
from unittest.mock import Mock


class ConfigItemTestCase(unittest.TestCase):
    def test_value(self):
        value = {'value': 'value', 'type': 'type'}
        item = ConfigItem(value)
        self.assertEqual(item.value, 'value')
    
    def test_change_prop(self):
        value = {'value': 'value', 'type': 'type'}
        item = ConfigItem(value)
        self.assertTrue(isinstance(item.change, Publisher))
    
    def test_change_is_called_when_value_is_change(self):
        value = {'value': 'value', 'type': 'type'}
        item = ConfigItem(value)
        test_dict = {}
        
        def func (currentVal, prevVal):
            test_dict['value'] = currentVal

        item.change.subscribe(func)
        item.value = 'change'
        self.assertEqual(test_dict['value'], item.value)

    def test_change_is_called_when_value_is_change_object(self):
        value = {'value': 'value', 'type': 'type'}
        item = ConfigItem(value)
        test_dict = {}
        
        def func (currentVal, prevVal):
            test_dict['value'] = currentVal

        item.change.subscribe(func)
        item.value = {'value': 'change', 'type': 'type'}
        self.assertEqual(test_dict['value'], item.value)

    def test_change_is_called_when_type_is_change_object(self):
        value = {'value': 'value', 'type': 'type'}
        item = ConfigItem(value)
        test_dict = {}
        
        def func (currentVal, prevVale):
            test_dict['value'] = currentVal

        item.change.subscribe(func)
        item.value = {'value': 'value', 'type': 'change'}
        self.assertEqual(test_dict['value'], item.value)

    def test_change_is_not_called_when_value_is_not_change(self):
        value = {'value': 'value', 'type': 'type'}
        item = ConfigItem(value)
        f = Mock()

        item.change.subscribe(f)
        item.value = 'value'
        f.assert_not_called()
    
    

    

if __name__ == '__main__':
    unittest.main()




import uuid

class Publisher:
    
    def __init__(self):
        self.subscribers = {}

    def subscribe (self, func):
        key = str(uuid.uuid4())
        self.subscribers[key] = func
        return Subscription(self, key)
    
    def unsubscribe (self, key):
        self.subscribers.pop(key, None)
    
    def call (self, key):
        self.subscribers[key]()
    
    def publish (self, currentValue, prevValue):
        for key, callback in self.subscribers.items():
            callback(currentValue, prevValue)



class Subscription:
    def __init__(self, publisher, key):
        self.publisher = publisher
        self.key = key

    def unsubscribe(self):
        self.publisher.unsubscribe(self.key)
    
    def call(self):
        self.publisher.call(self.key)



from microkit_python_server_sdk.publisher import Publisher
class Item:
    
    def __init__(self, value, name):
        self._value = value['value']
        self.type = value['type']
        self.name = name
        self.change = Publisher()

    @property
    def value (self):
        if self.type == 'string':
            return str(self._value)
        elif self.type == 'number':
            return float(self._value)
        elif self.type == 'boolean':
            return  True if self._value == 'true' or self._value == True else False
       



    @value.setter
    def value(self, value):
        prev_value = self._value
        if isinstance(value, dict):
            have_been_changed = self._value != value["value"] or self.type != value["type"]
            self._value = value["value"]
            self.type = value["type"]

        else:
            have_been_changed = self._value != value
            self._value = value
        if (have_been_changed):
            self.change.publish(self._value, prev_value)

    def update(self, value):    
        self.value = value
    
    

    



class Conf:
    conf = {"version": "v1"}
    def __init__(self, options, user):
        self.conf['base_url'] = options['base_url'] if 'base_url' in options else 'sdk.microkit.app'
        self.conf['api_key'] = self.validate_require(options, 'api_key')
        self.conf['port'] = options['port'] if 'port' in options else 443
        self.conf['http'] = options['http'] if 'http' in options else False
        self.conf['polling_interval'] = options['polling_interval'] if 'polling_interval' in options else 30
        self.conf['service'] = options['service'] if 'service' in options else ''
        self.conf['polling_on'] = options['polling_on'] if 'polling_on' in options else True
        self.conf['user'] = user

    def get (self, conf_name):
        if conf_name in self.conf:
            return self.conf[conf_name]
        else:
            return None

    def set (self, key, value):
        self.conf[key] = value
        
    def validate_require(self, options, value_name):
        if value_name in options:
            return options[value_name]
        else:
            raise Exception("{} is required.".format(value_name))
    

    



from microkit_python_server_sdk.http_client import HttpClient
from microkit_python_server_sdk.publisher import Publisher
import threading
import time

class NetworkController:
    
    def __init__(self, config):
        self.config = config
        self.client = HttpClient(config)
        self.change = Publisher()

    def get_latest_data (self, action):
        return self.client.post(action, None, {"user": self.config.get("user")})

    
    def start_update_interval (self):
        self.thread = StoppableThread(target=self.update)
        self.thread.start()

    def stop_update_interval (self):
        self.thread.stop()

    def update(self):
        count = 0
        while True:
            count = count + 1
            if self.thread.stopped():
                break
            if count == self.config.get('polling_interval'):
                count = 0
                res = self.get_latest_data('update')
                self.change.publish(res, {})
            time.sleep(1)



class StoppableThread(threading.Thread):

    def __init__(self,  *args, **kwargs):
        super(StoppableThread, self).__init__(*args, **kwargs)
        self._stop_event = threading.Event()

    def stop(self):
        self._stop_event.set()

    def stopped(self):
        return self._stop_event.is_set()


import unittest
from http_client import HttpClient
from conf import Conf
from unittest.mock import patch, MagicMock
from network_controller import NetworkController

data = {"key1": "value1"}

class NetworkControllerTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        config = Conf(api_key = 'key')
        # self.client = HttpClient(config)
        self.controller = NetworkController(config)
    
    def test_config(self):
        self.assertEqual(self.controller.config.get('api_key'), 'key')
    
    def test_client(self):
        self.assertTrue(isinstance(self.controller.client, HttpClient))
    
    @patch('urllib.request.urlopen')
    def test_get_latest_data(self, mock_urlopen):
        cm = MagicMock()
        cm.getcode.return_value = 200
        cm.read.return_value = data
        cm.__enter__.return_value = cm
        mock_urlopen.return_value = cm
       
        res = self.controller.get_latest_data()
        self.assertEqual(res, data)

    # test get 
    # test port
    # test http/https
    # test start with http
    # test valid json

if __name__ == '__main__':
    unittest.main()